-------------------------------------------------------------------------------
--
-- Copyright (C) 2009
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

PACKAGE unb1_board_pkg IS

  -- UniBoard
  CONSTANT c_unb1_board_nof_node             : NATURAL := 4;                            -- nof FN or nof BN on UniBoard
  CONSTANT c_unb1_board_nof_node_w           : NATURAL := 2;                            -- = ceil_log2(c_unb1_board_nof_node)
  CONSTANT c_unb1_board_nof_fn               : NATURAL := c_unb1_board_nof_node;               -- nof Front Node FPGAs on UniBoard
  CONSTANT c_unb1_board_nof_bn               : NATURAL := c_unb1_board_nof_node;               -- nof Back Node FPGAs on UniBoard
  CONSTANT c_unb1_board_nof_chip             : NATURAL := c_unb1_board_nof_fn + c_unb1_board_nof_bn;  -- = 8
  CONSTANT c_unb1_board_nof_chip_w           : NATURAL := 3;                            -- = ceil_log2(c_unb1_board_nof_chip)
  CONSTANT c_unb1_board_nof_ddr3             : NATURAL := 2;                            -- each node has 2 DDR3 modules
  
  -- Subrack
  CONSTANT c_unb1_board_nof_uniboard         : NATURAL := 4;   -- nof UniBoard in a subrack
  CONSTANT c_unb1_board_nof_uniboard_w       : NATURAL := 5;   -- Only 2 required for 4 boards; full width is 5.
  
  -- Clock frequencies
  CONSTANT c_unb1_board_ext_clk_freq_200M    : NATURAL := 200 * 10**6;  -- external clock, SMA clock
  CONSTANT c_unb1_board_eth_clk_freq         : NATURAL :=  25 * 10**6;  -- fixed 25 MHz ETH XO clock used as reference clock for the PLL in SOPC
  CONSTANT c_unb1_board_tse_clk_freq         : NATURAL := 125 * 10**6;  -- fixed 125 MHz TSE reference clock derived from ETH_clk by PLL in SOPC
  CONSTANT c_unb1_board_cal_clk_freq         : NATURAL :=  40 * 10**6;  -- fixed 40 MHz IO calibration clock derived from ETH_clk by PLL in SOPC
  CONSTANT c_unb1_board_mm_clk_freq_50M      : NATURAL :=  50 * 10**6;  -- SOPC memory mapped bus clock derived from ETH_clk by PLL in SOPC
  CONSTANT c_unb1_board_mm_clk_freq_100M     : NATURAL := 100 * 10**6;  -- SOPC memory mapped bus clock derived from ETH_clk by PLL in SOPC
  CONSTANT c_unb1_board_mm_clk_freq_125M     : NATURAL := 125 * 10**6;  -- SOPC memory mapped bus clock derived from ETH_clk by PLL in SOPC
  
  -- PHY interface device numbers
  CONSTANT c_unb1_board_phy_id_eth1g         : NATURAL := 1;
  CONSTANT c_unb1_board_phy_id_eth10g        : NATURAL := 2;
  CONSTANT c_unb1_board_phy_id_tr_mesh       : NATURAL := 3;
  CONSTANT c_unb1_board_phy_id_tr_back       : NATURAL := 4;
  CONSTANT c_unb1_board_phy_id_ddr3_I        : NATURAL := 5;
  CONSTANT c_unb1_board_phy_id_ddr3_II       : NATURAL := 6;
  CONSTANT c_unb1_board_phy_id_adc           : NATURAL := 7;
  
  -- PHY interface error numbers
  CONSTANT c_unb1_board_ok                   : NATURAL := 0;  -- value
  CONSTANT c_unb1_board_error                : INTEGER := 1;  -- value
  CONSTANT c_unb1_board_error_eth1g_bi       : NATURAL := 1;  -- = c_unb1_board_phy_id_eth1g;    -- bit number index
  CONSTANT c_unb1_board_error_eth10g_bi      : NATURAL := 2;  -- = c_unb1_board_phy_id_eth10g;
  CONSTANT c_unb1_board_error_tr_mesh_bi     : NATURAL := 3;  -- = c_unb1_board_phy_id_tr_mesh;
  CONSTANT c_unb1_board_error_tr_back_bi     : NATURAL := 4;  -- = c_unb1_board_phy_id_tr_back;
  CONSTANT c_unb1_board_error_ddr3_I_bi      : NATURAL := 5;  -- = c_unb1_board_phy_id_ddr3_I;
  CONSTANT c_unb1_board_error_ddr3_II_bi     : NATURAL := 6;  -- = c_unb1_board_phy_id_ddr3_II;
  CONSTANT c_unb1_board_error_adc_bi         : NATURAL := 7;  -- = c_unb1_board_phy_id_adc;
  CONSTANT c_unb1_board_error_w              : NATURAL := 8;
  
  -- I2C
  CONSTANT c_unb1_board_reg_sens_adr_w       : NATURAL := 3;  -- must match ceil_log2(c_mm_nof_dat) in unb1_board_sens_reg.vhd
  
  -- CONSTANT RECORD DECLARATIONS ---------------------------------------------
  
  -- PHY interface instantiation control
  TYPE t_c_unb1_board_use_phy IS RECORD
    eth1g    : NATURAL;
    tr_front : NATURAL;
    tr_mesh  : NATURAL;
    tr_back  : NATURAL;
    ddr3_I   : NATURAL;
    ddr3_II  : NATURAL;
    adc      : NATURAL;
    wdi      : NATURAL;  -- enable WDI when 1, else disable by leaving unconnected so 'Z'
  END RECORD;
  
  -- c_unb1_board_use_phy* : when 1 and applicable then the PHY module is instantiated in the design
  CONSTANT c_unb1_board_use_phy_none          : t_c_unb1_board_use_phy := (0, 0, 0, 0, 0, 0, 0, 0);
  CONSTANT c_unb1_board_use_phy_fn            : t_c_unb1_board_use_phy := (1, 1, 1, 0, 1, 1, 0, 1);
  CONSTANT c_unb1_board_use_phy_fn_no_wdi     : t_c_unb1_board_use_phy := (1, 1, 1, 0, 1, 1, 0, 0);
  CONSTANT c_unb1_board_use_phy_bn            : t_c_unb1_board_use_phy := (1, 0, 1, 1, 1, 1, 1, 1);
  CONSTANT c_unb1_board_use_phy_bn_no_wdi     : t_c_unb1_board_use_phy := (1, 0, 1, 1, 1, 1, 1, 0);
  CONSTANT c_unb1_board_use_phy_xgb           : t_c_unb1_board_use_phy := (1, 0, 1, 1, 1, 1, 0, 1);
  CONSTANT c_unb1_board_use_phy_xgb_no_wdi    : t_c_unb1_board_use_phy := (1, 0, 1, 1, 1, 1, 0, 0);
  CONSTANT c_unb1_board_use_phy_common        : t_c_unb1_board_use_phy := (1, 0, 1, 0, 1, 1, 0, 1);
  CONSTANT c_unb1_board_use_phy_common_no_wdi : t_c_unb1_board_use_phy := (1, 0, 1, 0, 1, 1, 0, 0);
  
  -- c_unb1_board_signature_* : random signature words used for unused status bits to ensure that the software reads the correct interface address
  CONSTANT c_unb1_board_signature_eth1g_slv   : STD_LOGIC_VECTOR(31 DOWNTO 0) := X"46e46cbc";
  CONSTANT c_unb1_board_signature_eth10g_slv  : STD_LOGIC_VECTOR(31 DOWNTO 0) := X"2bd2e40a";
  CONSTANT c_unb1_board_signature_tr_mesh_slv : STD_LOGIC_VECTOR(31 DOWNTO 0) := X"f702ef5f";  -- use TO_SINT to avoid integrer overflow warning in Quartus
  CONSTANT c_unb1_board_signature_tr_back_slv : STD_LOGIC_VECTOR(31 DOWNTO 0) := X"6bf88c24";
  CONSTANT c_unb1_board_signature_ddr3_I_slv  : STD_LOGIC_VECTOR(31 DOWNTO 0) := X"d96089c7";  -- use TO_SINT to avoid integrer overflow warning in Quartus
  CONSTANT c_unb1_board_signature_ddr3_II_slv : STD_LOGIC_VECTOR(31 DOWNTO 0) := X"cace197c";  -- use TO_SINT to avoid integrer overflow warning in Quartus

  CONSTANT c_unb1_board_signature_eth1g       : INTEGER := TO_SINT(c_unb1_board_signature_eth1g_slv  );
  CONSTANT c_unb1_board_signature_eth10g      : INTEGER := TO_SINT(c_unb1_board_signature_eth10g_slv );
  CONSTANT c_unb1_board_signature_tr_mesh     : INTEGER := TO_SINT(c_unb1_board_signature_tr_mesh_slv);
  CONSTANT c_unb1_board_signature_tr_back     : INTEGER := TO_SINT(c_unb1_board_signature_tr_back_slv);
  CONSTANT c_unb1_board_signature_ddr3_I      : INTEGER := TO_SINT(c_unb1_board_signature_ddr3_I_slv );
  CONSTANT c_unb1_board_signature_ddr3_II     : INTEGER := TO_SINT(c_unb1_board_signature_ddr3_II_slv);
  
  -- Transceivers
  TYPE t_c_unb1_board_tr IS RECORD
    nof_bus                           : NATURAL;  -- = 4;   --    FN-BN, BN-BI
    bus_w                             : NATURAL;  -- = 4;   --    FN-BN, BN-BI
    cntrl_w                           : NATURAL;  -- = 3;   -- SI-FN
    cntrl_lasi_id                     : NATURAL;  -- = 0;   -- SI-FN
    cntrl_mdc_id                      : NATURAL;  -- = 1;   -- SI-FN
    cntrl_mdio_id                     : NATURAL;  -- = 2;   -- SI-FN
  END RECORD;
  
  CONSTANT c_unb1_board_tr                   : t_c_unb1_board_tr := ( 4, 4, 3, 0, 1, 2);  -- UniBoard hardware has 4 bundles of 4 transceivers or 16 individual transceivers (including CMU transceivers)
  CONSTANT c_unb1_board_tr_mesh              : t_c_unb1_board_tr := ( 4, 3, 3, 0, 1, 2);  -- the FN-BN mesh   uses 4 bundles of 3 transceivers or 12 individual transceivers (not using the CMU transceivers)
  CONSTANT c_unb1_board_tr_back              : t_c_unb1_board_tr := ( 3, 4, 3, 0, 1, 2);  -- the BN-BI back   uses 3 bundles of 4 transceivers or 12 individual transceivers (not using the CMU transceivers)
  CONSTANT c_unb1_board_tr_xaui              : t_c_unb1_board_tr := ( 3, 4, 3, 0, 1, 2);  -- FN-XAUI          use  3 bundles of 4 transceivers (the CMU transceivers are not HW connected) and XAUI PHY control for 10GbE
  
  -- Transceivers network array types for the UniBoard mesh HW and for a backplane HW
  CONSTANT c_unb1_board_tr_mesh_hw_nof_bus   : NATURAL := c_unb1_board_nof_node;        -- fixed 4 FN and 4 BN on one UniBoard
  CONSTANT c_unb1_board_tr_mesh_hw_bus_w     : NATURAL := 4;                     -- effectively 4-1 = 3, because the CMU tranceivers are connected on the mesh, but typically not used
  CONSTANT c_unb1_board_tr_back_hw_nof_bus   : NATURAL := c_unb1_board_nof_uniboard;    -- assume 4 UniBoard in a subrack (could be more), so connect to 4 - 1= 3 other UniBoards in subrack. Also allow complete range 3..0 to be used for logical indexing (including the hosting board itself).
  CONSTANT c_unb1_board_tr_back_hw_bus_w     : NATURAL := 4;                     -- the CMU tranceivers are not connected on the backplane
  CONSTANT c_unb1_board_tr_xaui_hw_bus_w     : NATURAL := 4; 
  
  -- TR mesh node IO for 1 to 4 nodes
  TYPE t_unb1_board_mesh_sosi_2arr IS ARRAY (c_unb1_board_tr_mesh_hw_nof_bus-1 DOWNTO 0) OF t_dp_sosi_arr(   c_unb1_board_tr_mesh_hw_bus_w-1 DOWNTO 0);  -- indexing (node id 3,2,1,0)(transceiver lane 3,2,1,0)
  TYPE t_unb1_board_mesh_siso_2arr IS ARRAY (c_unb1_board_tr_mesh_hw_nof_bus-1 DOWNTO 0) OF t_dp_siso_arr(   c_unb1_board_tr_mesh_hw_bus_w-1 DOWNTO 0);  -- indexing (node id 3,2,1,0)(transceiver lane 3,2,1,0)
  TYPE t_unb1_board_mesh_sl_2arr   IS ARRAY (c_unb1_board_tr_mesh_hw_nof_bus-1 DOWNTO 0) OF STD_LOGIC_VECTOR(c_unb1_board_tr_mesh_hw_bus_w-1 DOWNTO 0);  -- indexing (node id 3,2,1,0)(transceiver lane 3,2,1,0)
  
  CONSTANT c_unb1_board_mesh_sosi_2arr_rst   : t_unb1_board_mesh_sosi_2arr := (OTHERS=>(OTHERS=>c_dp_sosi_rst));
  CONSTANT c_unb1_board_mesh_siso_2arr_rst   : t_unb1_board_mesh_siso_2arr := (OTHERS=>(OTHERS=>c_dp_siso_rst));
  CONSTANT c_unb1_board_mesh_sl_2arr_rst     : t_unb1_board_mesh_sl_2arr   := (OTHERS=>(OTHERS=>'0'));
  
  -- TR backplane node IO for 1 BN to 3 BN each on the 3 other UniBoards
  TYPE t_unb1_board_back_sosi_2arr IS ARRAY (c_unb1_board_tr_back_hw_nof_bus-1 DOWNTO 0) OF t_dp_sosi_arr(   c_unb1_board_tr_back_hw_bus_w-1 DOWNTO 0);  -- indexing (other uniboard 2,1,0)(transceiver lane 3,2,1,0) or indexing (uniboard 3,2,1,0)(transceiver lane 3,2,1,0)
  TYPE t_unb1_board_back_siso_2arr IS ARRAY (c_unb1_board_tr_back_hw_nof_bus-1 DOWNTO 0) OF t_dp_siso_arr(   c_unb1_board_tr_back_hw_bus_w-1 DOWNTO 0);  -- indexing (other uniboard 2,1,0)(transceiver lane 3,2,1,0) or indexing (uniboard 3,2,1,0)(transceiver lane 3,2,1,0)
  TYPE t_unb1_board_back_sl_2arr   IS ARRAY (c_unb1_board_tr_back_hw_nof_bus-1 DOWNTO 0) OF STD_LOGIC_VECTOR(c_unb1_board_tr_back_hw_bus_w-1 DOWNTO 0);  -- indexing (other uniboard 2,1,0)(transceiver lane 3,2,1,0)

  TYPE t_unb1_board_xaui_sl_2arr   IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(c_unb1_board_tr_xaui_hw_bus_w-1 DOWNTO 0); 

  CONSTANT c_unb1_board_back_sosi_2arr_rst   : t_unb1_board_back_sosi_2arr := (OTHERS=>(OTHERS=>c_dp_sosi_rst));
  CONSTANT c_unb1_board_back_siso_2arr_rst   : t_unb1_board_back_siso_2arr := (OTHERS=>(OTHERS=>c_dp_siso_rst));
  CONSTANT c_unb1_board_back_sl_2arr_rst     : t_unb1_board_back_sl_2arr   := (OTHERS=>(OTHERS=>'0'));
 
  -- DDR3 (definitions similar as in ug_altmemphy.pdf)
  TYPE t_c_unb1_board_ddr IS RECORD
    a_w                               : NATURAL;  -- = 16;
    ba_w                              : NATURAL;  -- = 3;
    dq_w                              : NATURAL;  -- = 64;
    nof_dq_per_dqs                    : NATURAL;  -- = 8;
    dqs_w                             : NATURAL;  -- = 8;  -- = dq_w / nof_dq_per_dqs;
    dm_w                              : NATURAL;  -- = 8;
    cs_w                              : NATURAL;  -- = 2;
    clk_w                             : NATURAL;  -- = 2;
  END RECORD;

  CONSTANT c_unb1_board_ddr                  : t_c_unb1_board_ddr := (16, 3, 64, 8, 8, 8, 2, 2);
  CONSTANT c_unb1_board_ddr_4g               : t_c_unb1_board_ddr := (15, 3, 64, 8, 8, 8, 2, 2);
  CONSTANT c_unb1_board_ddr_1g               : t_c_unb1_board_ddr := (14, 3, 64, 8, 8, 8, 1, 2);
  
  -- Auxiliary
  
  -- Test IO Interface
  TYPE t_c_unb1_board_testio IS RECORD  
    tst_w                             : NATURAL;  -- = nof tst = 4; [tst_w-1 +tst_lo : tst_lo] = [7:4],
    led_w                             : NATURAL;  -- = nof led = 2; [led_w-1 +led_lo : led_lo] = [3:2],
    jmp_w                             : NATURAL;  -- = nof jmp = 2; [jmp_w-1 +jmp_lo : jmp_lo] = [1:0],
    tst_lo                            : NATURAL;  -- = 4;
    led_lo                            : NATURAL;  -- = 2;
    jmp_lo                            : NATURAL;  -- = 0;
  END RECORD;

  CONSTANT c_unb1_board_testio               : t_c_unb1_board_testio := (4, 2, 2, 4, 2, 0);
  CONSTANT c_unb1_board_testio_led_green     : NATURAL := c_unb1_board_testio.led_lo;
  CONSTANT c_unb1_board_testio_led_red       : NATURAL := c_unb1_board_testio.led_lo+1;
  
  TYPE t_c_unb1_board_aux IS RECORD
    version_w                         : NATURAL;  -- = 2;
    id_w                              : NATURAL;  -- = 8;  -- 5+3 bits wide = total node ID for up to 32 UniBoards in a system and 8 nodes per board
    chip_id_w                         : NATURAL;  -- = 3;  -- board node ID for the 8 FPGA nodes on a UniBoard
    testio_w                          : NATURAL;  -- = 8;
    testio                            : t_c_unb1_board_testio;
  END RECORD;
  
  CONSTANT c_unb1_board_aux                  : t_c_unb1_board_aux := (2, 8, c_unb1_board_nof_chip_w, 8, c_unb1_board_testio);
  
  TYPE t_e_unb1_board_node IS (e_fn, e_bn, e_any);

  -- UniBoard Common Interface
  TYPE t_c_unb1_board_ci IS RECORD
    tr                                : t_c_unb1_board_tr;
    ddr                               : t_c_unb1_board_ddr;
    aux                               : t_c_unb1_board_aux;
  END RECORD;
  
  CONSTANT c_unb1_board_ci                   : t_c_unb1_board_ci := (c_unb1_board_tr, c_unb1_board_ddr, c_unb1_board_aux);

  -- ADC Interface
  TYPE t_c_unb1_board_ai IS RECORD
    nof_ports                         : NATURAL;  -- = 4;     -- Fixed 4 ADC BI port with names A, B, C, D
    nof_ovr                           : NATURAL;  -- = 2;     -- Fixed 2 overflow signals, one for ports AB and one for ports CD
    port_w                            : NATURAL;  -- = 8;     -- Fixed 8 bit ADC BI port width, the ADC sample width is also 8 bit
    lvds_data_rate                    : NATURAL;  -- = 800;   -- The ADC sample rate is 800 Msps, so the LVDS rate is 800 Mbps per ADC BI data line, 
    use_dpa                           : BOOLEAN;  -- = TRUE;  -- When TRUE use LVDS_RX with DPA, else used fixed IOE delays and/or lvds_clk_phase instead of DPA
    use_lvds_clk                      : BOOLEAN;  -- = TRUE;  -- When TRUE use the one or both ADC BI lvds_clk, else use the single dp_clk to capture the lvds data
    use_lvds_clk_rst                  : BOOLEAN;  -- = FALSE; -- When TRUE then support reset pulse to ADU to align the lvds_clk to the dp_clk, else no support
    lvds_clk_phase                    : NATURAL;  -- = 0;     -- Use PLL phase 0 for center aligned. Only for no DPA
    nof_clocks                        : NATURAL;  -- = 2;     -- 1 --> Use ADC BI clock D or dp_clk and 32 bit port ABCD
                                                              -- 2 --> Use ADC BI clock A, D and 16 bit ports AB, CD
    lvds_deser_factor                 : NATURAL;  -- = 2;     -- The ADC sampled data comes in with a DDR lvds_clk, so lvds_data_rate / 2
    dp_deser_factor                   : NATURAL;  -- = 4;     -- The Data Path clock dp_clk frequency is 200 MHz, so lvds_data_rate / 4
  END RECORD;
  
  CONSTANT c_unb1_board_ai                   : t_c_unb1_board_ai := (4, 2, 8, 800, TRUE, TRUE, FALSE, 0, 2, 2, 4);
  
  TYPE t_unb1_board_fw_version IS RECORD
    hi                                : NATURAL;  -- = 0..15
    lo                                : NATURAL;  -- = 0..15, firmware version is: hi.lo
  END RECORD;
  
  CONSTANT c_unb1_board_fw_version           : t_unb1_board_fw_version := (0, 0);
    
  -- SIGNAL RECORD DECLARATIONS -----------------------------------------------
  
  -- DDR3 (definitions similar as in ug_altmemphy.pdf)
  TYPE t_unb1_board_ddr_in IS RECORD  
    evt              : STD_LOGIC;
    --nc               : STD_LOGIC;   -- not connect, needed to be able to initialize constant record which has to have more than one field in VHDL
  END RECORD;
  
  TYPE t_unb1_board_ddr_inout IS RECORD  
    dq               : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.dq_w-1 DOWNTO 0);   -- data bus
    dqs              : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.dqs_w-1 DOWNTO 0);  -- data strobe bus
    dqs_n            : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.dqs_w-1 DOWNTO 0);
    clk              : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.clk_w-1 DOWNTO 0);  -- clock, positive edge clock
    clk_n            : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.clk_w-1 DOWNTO 0);  -- clock, negative edge clock
    scl              : STD_LOGIC;                                        -- I2C
    sda              : STD_LOGIC;
  END RECORD;
  
  TYPE t_unb1_board_ddr_out IS RECORD  
    a                : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.a_w-1 DOWNTO 0);    -- row and column address
    ba               : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.ba_w-1 DOWNTO 0);   -- bank address
    dm               : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.dm_w-1 DOWNTO 0);   -- data mask bus
    cas_n            : STD_LOGIC;                                        -- column address strobe
    ras_n            : STD_LOGIC;                                        -- row address strobe
    we_n             : STD_LOGIC;                                        -- write enable signal
    reset_n          : STD_LOGIC;                                        -- reset signal
    odt              : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.cs_w-1 DOWNTO 0);   -- on-die termination control signal
    cke              : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.cs_w-1 DOWNTO 0);   -- clock enable
    cs_n             : STD_LOGIC_VECTOR(c_unb1_board_ci.ddr.cs_w-1 DOWNTO 0);   -- chip select
  END RECORD;

  --CONSTANT c_unb1_board_ddr_in_rst    : t_unb1_board_ddr_in    := ('0', 'X');
  CONSTANT c_unb1_board_ddr_inout_rst : t_unb1_board_ddr_inout := ((OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), '0', '0');
  CONSTANT c_unb1_board_ddr_out_rst   : t_unb1_board_ddr_out   := ((OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), '0', '0', '0', '0', (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'));
  
  -- I2C, MDIO
  -- . If no I2C bus arbitration or clock stretching is needed then the SCL only needs to be output.
  -- . Can also be used for a PHY Management Data IO interface with serial clock MDC and serial data MDIO
  TYPE t_unb1_board_i2c_inout IS RECORD  
    scl : STD_LOGIC;  -- serial clock
    sda : STD_LOGIC;  -- serial data
  END RECORD;
    
  -- System info
  TYPE t_c_unb1_board_system_info IS RECORD
    version  : NATURAL;  -- UniBoard board HW version (2 bit value)
    id       : NATURAL;  -- UniBoard FPGA node id (8 bit value)
                         -- Derived ID info:
    bck_id   : NATURAL;  -- = id[7:3], ID part from back plane
    chip_id  : NATURAL;  -- = id[2:0], ID part from UniBoard
    is_bn    : NATURAL;  -- = id[2], 0 for Front Node, 1 for Back Node
    node_id  : NATURAL;  -- = id[1:0], node ID: 0, 1, 2 or 3
    is_bn3   : NATURAL;  -- 1 for Back Node 3, else 0.
  END RECORD;

  FUNCTION func_unb1_board_system_info(VERSION : IN STD_LOGIC_VECTOR(c_unb1_board_aux.version_w-1 DOWNTO 0);
                                       ID      : IN STD_LOGIC_VECTOR(c_unb1_board_aux.id_w-1 DOWNTO 0)) RETURN t_c_unb1_board_system_info;
                                
  FUNCTION func_unb1_board_chip_id(chip_id   : IN STD_LOGIC_VECTOR(c_unb1_board_aux.chip_id_w-1 DOWNTO 0);
                                   node_type : IN t_e_unb1_board_node) RETURN STD_LOGIC_VECTOR;
  
  -- Connect: out_2arr = in_2arr of different types
  FUNCTION func_unb1_board_connect_2arr(in_2arr : t_unb1_board_mesh_sosi_2arr) RETURN t_unb1_board_back_sosi_2arr;
  FUNCTION func_unb1_board_connect_2arr(in_2arr : t_unb1_board_mesh_siso_2arr) RETURN t_unb1_board_back_siso_2arr;
  FUNCTION func_unb1_board_connect_2arr(in_2arr : t_unb1_board_back_sosi_2arr) RETURN t_unb1_board_mesh_sosi_2arr;
  FUNCTION func_unb1_board_connect_2arr(in_2arr : t_unb1_board_back_siso_2arr) RETURN t_unb1_board_mesh_siso_2arr;
  
  -- Transpose: out_2arr(J)(I) = in_2arr(I)(J) for different types
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_mesh_sosi_2arr) RETURN t_unb1_board_back_sosi_2arr;
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_mesh_siso_2arr) RETURN t_unb1_board_back_siso_2arr;
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_back_sosi_2arr) RETURN t_unb1_board_mesh_sosi_2arr;
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_back_siso_2arr) RETURN t_unb1_board_mesh_siso_2arr;
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_mesh_sosi_2arr) RETURN t_unb1_board_mesh_sosi_2arr;
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_mesh_siso_2arr) RETURN t_unb1_board_mesh_siso_2arr;
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_back_sosi_2arr) RETURN t_unb1_board_back_sosi_2arr;
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_back_siso_2arr) RETURN t_unb1_board_back_siso_2arr;
  
END unb1_board_pkg;


PACKAGE BODY unb1_board_pkg IS

  FUNCTION func_unb1_board_system_info(VERSION : IN STD_LOGIC_VECTOR(c_unb1_board_aux.version_w-1 DOWNTO 0);
                                       ID      : IN STD_LOGIC_VECTOR(c_unb1_board_aux.id_w-1 DOWNTO 0)) RETURN t_c_unb1_board_system_info IS
    VARIABLE v_system_info : t_c_unb1_board_system_info;
  BEGIN
    v_system_info.version := TO_INTEGER(UNSIGNED(VERSION));
    v_system_info.id      := TO_INTEGER(UNSIGNED(ID));
    v_system_info.bck_id  := TO_INTEGER(UNSIGNED(ID(7 DOWNTO 3)));
    v_system_info.chip_id := TO_INTEGER(UNSIGNED(ID(2 DOWNTO 0)));
    IF ID(2)='1' THEN v_system_info.is_bn := 1; ELSE v_system_info.is_bn := 0; END IF;
    v_system_info.node_id := TO_INTEGER(UNSIGNED(ID(1 DOWNTO 0)));
    IF UNSIGNED(ID(2 DOWNTO 0))=7 THEN v_system_info.is_bn3 := 1; ELSE v_system_info.is_bn3 := 0; END IF;
    RETURN v_system_info;
  END;

  FUNCTION func_unb1_board_chip_id(chip_id   : IN STD_LOGIC_VECTOR(c_unb1_board_aux.chip_id_w-1 DOWNTO 0);
                                   node_type : IN t_e_unb1_board_node) RETURN STD_LOGIC_VECTOR IS
    VARIABLE v_chip_id : STD_LOGIC_VECTOR(chip_id'RANGE);  -- [2:0]
  BEGIN
    v_chip_id := chip_id;    -- default for design that can run on either FN or BN
    IF node_type=e_fn THEN
      v_chip_id(2) := '0';   -- design that is intended to run on a FN 
    ELSIF node_type=e_bn THEN
      v_chip_id(2) := '1';   -- design that is intended to run on a BN
    END IF;
    RETURN v_chip_id;
  END;
  
  
  ------------------------------------------------------------------------------
  -- Connect mesh 2arr - back 2arr
  ------------------------------------------------------------------------------
  
  FUNCTION func_unb1_board_connect_2arr(in_2arr : t_unb1_board_mesh_sosi_2arr) RETURN t_unb1_board_back_sosi_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_mesh_hw_nof_bus;  -- = 4 = c_unb1_board_tr_back_hw_nof_bus
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_mesh_hw_bus_w;    -- = 4 = c_unb1_board_tr_back_hw_bus_w
    
    VARIABLE v_out_2arr : t_unb1_board_back_sosi_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(I)(J) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;
  
  FUNCTION func_unb1_board_connect_2arr(in_2arr : t_unb1_board_mesh_siso_2arr) RETURN t_unb1_board_back_siso_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_mesh_hw_nof_bus;  -- = 4 = c_unb1_board_tr_back_hw_nof_bus
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_mesh_hw_bus_w;    -- = 4 = c_unb1_board_tr_back_hw_bus_w
    
    VARIABLE v_out_2arr : t_unb1_board_back_siso_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(I)(J) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;
  
  FUNCTION func_unb1_board_connect_2arr(in_2arr : t_unb1_board_back_sosi_2arr) RETURN t_unb1_board_mesh_sosi_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_back_hw_nof_bus;  -- = 4 = c_unb1_board_tr_mesh_hw_nof_bus
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_back_hw_bus_w;    -- = 4 = c_unb1_board_tr_mesh_hw_bus_w
    
    VARIABLE v_out_2arr : t_unb1_board_mesh_sosi_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(I)(J) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;
  
  FUNCTION func_unb1_board_connect_2arr(in_2arr : t_unb1_board_back_siso_2arr) RETURN t_unb1_board_mesh_siso_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_back_hw_nof_bus;  -- = 4 = c_unb1_board_tr_mesh_hw_nof_bus
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_back_hw_bus_w;    -- = 4 = c_unb1_board_tr_mesh_hw_bus_w
    
    VARIABLE v_out_2arr : t_unb1_board_mesh_siso_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(I)(J) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;
  
  ------------------------------------------------------------------------------
  -- Transpose mesh 2arr - back 2arr
  ------------------------------------------------------------------------------
  
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_mesh_sosi_2arr) RETURN t_unb1_board_back_sosi_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_mesh_hw_nof_bus;  -- = 4 = c_unb1_board_tr_back_hw_nof_bus
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_mesh_hw_bus_w;    -- = 4 = c_unb1_board_tr_back_hw_bus_w
    
    VARIABLE v_out_2arr : t_unb1_board_back_sosi_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(J)(I) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;
  
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_mesh_siso_2arr) RETURN t_unb1_board_back_siso_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_mesh_hw_nof_bus;  -- = 4 = c_unb1_board_tr_back_hw_nof_bus
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_mesh_hw_bus_w;    -- = 4 = c_unb1_board_tr_back_hw_bus_w
    
    VARIABLE v_out_2arr : t_unb1_board_back_siso_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(J)(I) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;
  
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_back_sosi_2arr) RETURN t_unb1_board_mesh_sosi_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_back_hw_nof_bus;  -- = 4 = c_unb1_board_tr_mesh_hw_nof_bus
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_back_hw_bus_w;    -- = 4 = c_unb1_board_tr_mesh_hw_bus_w
    
    VARIABLE v_out_2arr : t_unb1_board_mesh_sosi_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(J)(I) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;
  
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_back_siso_2arr) RETURN t_unb1_board_mesh_siso_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_back_hw_nof_bus;  -- = 4 = c_unb1_board_tr_mesh_hw_nof_bus
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_back_hw_bus_w;    -- = 4 = c_unb1_board_tr_mesh_hw_bus_w
    
    VARIABLE v_out_2arr : t_unb1_board_mesh_siso_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(J)(I) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;

  ------------------------------------------------------------------------------
  -- Transpose mesh 2arr -mesh 2arr
  ------------------------------------------------------------------------------

  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_mesh_sosi_2arr) RETURN t_unb1_board_mesh_sosi_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_mesh_hw_nof_bus;  -- = 4 
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_mesh_hw_bus_w;    -- = 4
    
    VARIABLE v_out_2arr : t_unb1_board_mesh_sosi_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(J)(I) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;
  
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_mesh_siso_2arr) RETURN t_unb1_board_mesh_siso_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_mesh_hw_nof_bus;  -- = 4
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_mesh_hw_bus_w;    -- = 4
    
    VARIABLE v_out_2arr : t_unb1_board_mesh_siso_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(J)(I) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;

  ------------------------------------------------------------------------------
  -- Transpose back 2arr -back 2arr
  ------------------------------------------------------------------------------

  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_back_sosi_2arr) RETURN t_unb1_board_back_sosi_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_back_hw_nof_bus;  -- = 4
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_back_hw_bus_w;    -- = 4
    
    VARIABLE v_out_2arr : t_unb1_board_back_sosi_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(J)(I) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;
  
  FUNCTION func_unb1_board_transpose_2arr(in_2arr : t_unb1_board_back_siso_2arr) RETURN t_unb1_board_back_siso_2arr IS
    CONSTANT c_nof_bus : NATURAL := c_unb1_board_tr_back_hw_nof_bus;  -- = 4 
    CONSTANT c_bus_w   : NATURAL := c_unb1_board_tr_back_hw_bus_w;    -- = 4
    
    VARIABLE v_out_2arr : t_unb1_board_back_siso_2arr;
  BEGIN
    FOR I IN c_nof_bus-1 DOWNTO 0 LOOP
      FOR J IN c_bus_w-1 DOWNTO 0 LOOP
        v_out_2arr(J)(I) := in_2arr(I)(J);
      END LOOP;
    END LOOP;
    RETURN v_out_2arr;
  END;  
END unb1_board_pkg;
