-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;

-- Purpose:
--   Provide the basic node control along with an SOPC Builder system:
--   . sys_rst for sys_clk
--   . pulse every 1 us, 1 ms and 1 s
--   . extend WDI to avoid watchdog reset during SW reload

ENTITY unb2_board_node_ctrl IS
  GENERIC (
    g_pulse_us     : NATURAL := 125;     -- nof system clock cycles to get us period, equal to system clock frequency / 10**6
    g_pulse_ms     : NATURAL := 1000;    -- nof pulse_us pulses to get ms period (fixed, use less to speed up simulation)
    g_pulse_s      : NATURAL := 1000;    -- nof pulse_ms pulses to get  s period (fixed, use less to speed up simulation)
    g_wdi_extend_w : NATURAL := 14       -- extend wdi by about 2**(14-1)= 8 s
  );
  PORT (
    xo_clk          : IN  STD_LOGIC;  -- from pin, also used as reference for the PLL in the SOPC design
    xo_rst_n        : OUT STD_LOGIC;  -- to SOPC design
    sys_clk         : IN  STD_LOGIC;  -- system clock from PLL in SOPC design (= mm_clk)
    sys_locked      : IN  STD_LOGIC;  -- system clock PLL locked from SOPC design
    sys_rst         : OUT STD_LOGIC;  -- system reset released after system clock PLL has locked (= mm_rst)
    cal_clk         : IN  STD_LOGIC := '0';  -- calibration or configuration interface clock
    cal_rst         : OUT STD_LOGIC;         -- calibration or configuration interface reset, released after sys_rst is released
    st_clk          : IN  STD_LOGIC := '0';  -- streaming interface clock
    st_rst          : OUT STD_LOGIC;         -- streaming interface reset, released after sys_rst is released
    wdi_in          : IN  STD_LOGIC;  -- from SW running on the NIOS2 in the SOPC design
    wdi_out         : OUT STD_LOGIC;  -- to FPGA pin
    pulse_us        : OUT STD_LOGIC;  -- pulses every us
    pulse_ms        : OUT STD_LOGIC;  -- pulses every ms
    pulse_s         : OUT STD_LOGIC   -- pulses every s
  );
END unb2_board_node_ctrl;


ARCHITECTURE str OF unb2_board_node_ctrl IS

  CONSTANT c_reset_len   : NATURAL := 4;  -- >= c_meta_delay_len from common_pkg

  SIGNAL i_sys_rst   : STD_LOGIC;
  SIGNAL i_pulse_ms  : STD_LOGIC;
  
BEGIN

  sys_rst <= i_sys_rst;
  
  pulse_ms <= i_pulse_ms;

  u_unb2_board_clk_rst : ENTITY work.unb2_board_clk_rst
  PORT MAP (
    xo_clk     => xo_clk,
    xo_rst_n   => xo_rst_n,
    sys_clk    => sys_clk,
    sys_locked => sys_locked,
    sys_rst    => i_sys_rst   -- release reset some clock cycles after sys_locked went high
  );
  
  u_common_areset_cal : ENTITY common_lib.common_areset
  GENERIC MAP (
    g_rst_level => '1',
    g_delay_len => c_reset_len
  )
  PORT MAP (
    in_rst    => i_sys_rst,   -- release reset some clock cycles after i_sys_rst went low
    clk       => cal_clk,
    out_rst   => cal_rst
  );
  
  u_common_areset_st : ENTITY common_lib.common_areset
  GENERIC MAP (
    g_rst_level => '1',
    g_delay_len => c_reset_len
  )
  PORT MAP (
    in_rst    => i_sys_rst,   -- release reset some clock cycles after i_sys_rst went low
    clk       => st_clk,
    out_rst   => st_rst
  );
  
  u_common_pulser_us_ms_s : ENTITY common_lib.common_pulser_us_ms_s
  GENERIC MAP (
    g_pulse_us  => g_pulse_us,
    g_pulse_ms  => g_pulse_ms,
    g_pulse_s   => g_pulse_s
  )
  PORT MAP (
    rst         => i_sys_rst,
    clk         => sys_clk,
    pulse_us    => pulse_us,
    pulse_ms    => i_pulse_ms,
    pulse_s     => pulse_s
  );

  u_unb2_board_wdi_extend : ENTITY work.unb2_board_wdi_extend
  GENERIC MAP (
    g_extend_w => g_wdi_extend_w
  )
  PORT MAP (
    rst        => i_sys_rst,
    clk        => sys_clk, 
    pulse_ms   => i_pulse_ms,
    wdi_in     => wdi_in,
    wdi_out    => wdi_out
  );
  
END str;
