-------------------------------------------------------------------------------
--
-- Copyright (C) 2015
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Test bench for unb2_board_qsfp_leds
-- Description:
--   The test bench is self-stopping but not self-checking. Manually obeserve
--   in the wave window that:
--   1) factory image:
--      - green led is off
--      - red led toggles
--   2) user image
--      - red led is off
--      - green led toggles when any xon='0'
--      - green led is on continously when any xon='1'
--      - green led goes briefly off when any sop='1'
-- Usage:
--   > as 3
--   > run -a

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

ENTITY tb_unb2_board_qsfp_leds IS
END tb_unb2_board_qsfp_leds;

ARCHITECTURE tb OF tb_unb2_board_qsfp_leds IS

  CONSTANT c_clk_freq_hz    : NATURAL := 200 * 10**6;
  CONSTANT c_clk_period_ns  : NATURAL := 10**9 / c_clk_freq_hz;
  CONSTANT c_nof_clk_per_us : NATURAL := 1000 / c_clk_period_ns;
  
  CONSTANT clk_period       : TIME := c_clk_period_ns * 1 ns;
  
  CONSTANT c_nof_qsfp       : NATURAL := 2;
  CONSTANT c_nof_lanes      : NATURAL := c_nof_qsfp*c_quad;
  
  SIGNAL tb_end                  : STD_LOGIC := '0';
  SIGNAL rst                     : STD_LOGIC;
  SIGNAL clk                     : STD_LOGIC := '0';
  SIGNAL pulse_us                : STD_LOGIC;
  SIGNAL pulse_ms                : STD_LOGIC;
  SIGNAL pulse_s                 : STD_LOGIC;
 
  SIGNAL tx_siso_arr             : t_dp_siso_arr(c_nof_lanes-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rst);
  SIGNAL tx_sosi_arr             : t_dp_sosi_arr(c_nof_lanes-1 DOWNTO 0) := (OTHERS=>c_dp_sosi_rst);
  SIGNAL rx_sosi_arr             : t_dp_sosi_arr(c_nof_lanes-1 DOWNTO 0) := (OTHERS=>c_dp_sosi_rst);
    
  SIGNAL dbg_xon_arr             : STD_LOGIC_VECTOR(c_nof_lanes-1 DOWNTO 0);
  SIGNAL dbg_tx_sop_arr          : STD_LOGIC_VECTOR(c_nof_lanes-1 DOWNTO 0);
  SIGNAL dbg_rx_sop_arr          : STD_LOGIC_VECTOR(c_nof_lanes-1 DOWNTO 0);
  
  SIGNAL factory_green_led_arr   : STD_LOGIC_VECTOR(c_nof_qsfp-1 DOWNTO 0);
  SIGNAL factory_red_led_arr     : STD_LOGIC_VECTOR(c_nof_qsfp-1 DOWNTO 0);
  
  SIGNAL user_green_led_arr      : STD_LOGIC_VECTOR(c_nof_qsfp-1 DOWNTO 0);
  SIGNAL user_red_led_arr        : STD_LOGIC_VECTOR(c_nof_qsfp-1 DOWNTO 0);
  
  -- Cannot use proc_common_gen_pulse() to create sop in array.
  -- proc_common_gen_pulse() works for dbg_sop, dbg_sosi.sop but not for dbg_sop_slv(I) or for tx_sosi_arr(I).sop.
  -- The compiler then gives Error: "(vcom-1450) Actual (indexed name) for formal "pulse" is not a static signal name"
  -- It does work if the array index is from a GENERATE statement, but it does not work when it is from a LOOP statement.
  SIGNAL dbg_sop      : STD_LOGIC;
  SIGNAL dbg_sop_slv  : STD_LOGIC_VECTOR(c_nof_lanes-1 DOWNTO 0);
  SIGNAL dbg_sosi : t_dp_sosi;
  
BEGIN
  
  clk <= NOT clk OR tb_end AFTER clk_period/2;
  rst <= '1', '0' AFTER clk_period*7;
  
  -- Ease observation of record fields in Wave window, by mapping them to a SLV
  dbg_xon_arr    <= func_dp_stream_arr_get(tx_siso_arr, "XON");
  dbg_tx_sop_arr <= func_dp_stream_arr_get(tx_sosi_arr, "SOP");
  dbg_rx_sop_arr <= func_dp_stream_arr_get(rx_sosi_arr, "SOP");
  
  p_stimuli : PROCESS
  BEGIN
    tx_siso_arr <= (OTHERS=>c_dp_siso_rst);
    tx_sosi_arr <= (OTHERS=>c_dp_sosi_rst);
    rx_sosi_arr <= (OTHERS=>c_dp_sosi_rst);
    proc_common_wait_some_pulses(clk, pulse_ms, 50);
    
    -- Switch on each lane
    FOR I IN 0 TO c_nof_lanes-1 LOOP
      tx_siso_arr(I).xon <= '1';
      proc_common_wait_some_pulses(clk, pulse_ms, 10);
    END LOOP;
    proc_common_wait_some_pulses(clk, pulse_ms, 50);
    
    -- Issue the sop of a Tx packet on each lane
    FOR I IN 0 TO c_nof_lanes-1 LOOP
      -- Cannot use proc_common_gen_pulse(), because index I in a LOOP is not static
      tx_sosi_arr(I).sop <= '1';
      WAIT UNTIL rising_edge(clk);
      tx_sosi_arr(I).sop <= '0';
      proc_common_wait_some_pulses(clk, pulse_ms, 10);
    END LOOP;
    proc_common_wait_some_pulses(clk, pulse_ms, 50);
    
    -- Issue the sop of an Rx packet on each lane
    FOR I IN 0 TO c_nof_lanes-1 LOOP
      -- Cannot use proc_common_gen_pulse(), because index I in a LOOP is not static
      rx_sosi_arr(I).sop <= '1';
      WAIT UNTIL rising_edge(clk);
      rx_sosi_arr(I).sop <= '0';
      proc_common_wait_some_pulses(clk, pulse_ms, 10);
    END LOOP;
    proc_common_wait_some_pulses(clk, pulse_ms, 50);
    
    -- Switch off each lane
    FOR I IN 0 TO c_nof_lanes-1 LOOP
      tx_siso_arr(I).xon <= '0';
      proc_common_wait_some_pulses(clk, pulse_ms, 10);
    END LOOP;
    proc_common_wait_some_pulses(clk, pulse_ms, 50);
    
    tb_end <= '1';
    proc_common_wait_some_pulses(clk, pulse_ms, 10);
    WAIT;
  END PROCESS;
  
  u_unb2_factory_qsfp_leds : ENTITY work.unb2_board_qsfp_leds
  GENERIC MAP (
    g_sim             => TRUE,             -- when true speed up led toggling in simulation
    g_factory_image   => TRUE,             -- distinguish factory image and user images
    g_nof_qsfp        => c_nof_qsfp,       -- number of QSFP cages each with one dual led that can light red or green (or amber = red + green)
    g_pulse_us        => c_nof_clk_per_us  -- nof clk cycles to get us period
  )
  PORT MAP (
    rst               => rst,
    clk               => clk,
    -- internal pulser outputs
    pulse_us          => pulse_us,
    pulse_ms          => pulse_ms,
    pulse_s           => pulse_s,
    -- lane status
    tx_siso_arr       => tx_siso_arr,
    tx_sosi_arr       => tx_sosi_arr,
    rx_sosi_arr       => rx_sosi_arr,
    -- leds
    green_led_arr     => factory_green_led_arr,
    red_led_arr       => factory_red_led_arr
  );
  
  u_unb2_user_qsfp_leds : ENTITY work.unb2_board_qsfp_leds
  GENERIC MAP (
    g_sim             => TRUE,             -- when true speed up led toggling in simulation
    g_factory_image   => FALSE,            -- distinguish factory image and user images
    g_nof_qsfp        => c_nof_qsfp,       -- number of QSFP cages each with one dual led that can light red or green (or amber = red + green)
    g_pulse_us        => c_nof_clk_per_us  -- nof clk cycles to get us period
  )
  PORT MAP (
    rst               => rst,
    clk               => clk,
    -- internal pulser outputs
    pulse_us          => pulse_us,
    pulse_ms          => pulse_ms,
    pulse_s           => pulse_s,
    -- lane status
    tx_siso_arr       => tx_siso_arr,
    tx_sosi_arr       => tx_sosi_arr,
    rx_sosi_arr       => rx_sosi_arr,
    -- leds
    green_led_arr     => user_green_led_arr,
    red_led_arr       => user_red_led_arr
  );
  
END tb;
