 -------------------------------------------------------------------------------
--
-- Copyright (C) 2015
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Provide MM access via slave register to diag_tx_seq
-- Description:
--
--   Each DP stream has its own diag_tx_seq, because each stream can have its
--   own flow control.
--
--   One MM control register sets all g_nof_streams:
--
--   31             24 23             16 15              8 7               0  wi
--  |-----------------|-----------------|-----------------|-----------------|
--  |                                            select = [1], enable = [0] |  0
--  |-----------------------------------------------------------------------|
--  |                                   init[31: 0]                         |  1
--  |-----------------------------------------------------------------------|
--
-- . g_seq_dat_w
--   The g_seq_dat_w must be >= 1. The DP streaming data field is
--   c_dp_stream_data_w bits wide and the REPLICATE_DP_DATA() is used to wire
--   the g_seq_dat_w from the u_diag_tx_seq to fill the entire DP data width.
--   The maximum g_seq_dat_w depends on the pseudo random data width of the
--   LFSR sequeces in common_lfsr_sequences_pkg and on whether timing closure
--   can still be achieved for wider g_seq_dat_w. Thanks to the replication a
--   smaller g_seq_dat_w can be used to provide CNTR or LFSR data for the DP
--   data.
--
-- . diag_init
--   Note that MM diag_init has c_word_w=32 bits, so if g_seq_dat_w is wider
--   then the MSbits are 0 and if it is smaller, then the MSbits are ignored.

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL; 
USE common_lib.common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

ENTITY mms_diag_tx_seq IS
  GENERIC (
    g_nof_streams : NATURAL := 1;
    g_seq_dat_w   : NATURAL := c_word_w  -- >= 1, test sequence data width
  );
  PORT (
    -- Clocks and reset
    mm_rst         : IN  STD_LOGIC;  -- reset synchronous with mm_clk
    mm_clk         : IN  STD_LOGIC;  -- MM bus clock
    dp_rst         : IN  STD_LOGIC;  -- reset synchronous with dp_clk
    dp_clk         : IN  STD_LOGIC;  -- DP streaming bus clock

    -- MM interface
    reg_mosi       : IN  t_mem_mosi;   -- single MM control register applied to all g_nof_streams
    reg_miso       : OUT t_mem_miso;

    -- DP streaming interface
    tx_src_out_arr : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
    tx_src_in_arr  : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0)
  );
END mms_diag_tx_seq;


ARCHITECTURE str OF mms_diag_tx_seq IS    
  
  -- Define the actual size of the MM slave register
  CONSTANT c_mm_reg      : t_c_mem  := (latency  => 1,
                                        adr_w    => 1,
                                        dat_w    => c_word_w,       -- Use MM bus data width = c_word_w = 32 for all MM registers
                                        nof_dat  => 2,
                                        init_sl  => '0');

  TYPE t_dat_arr IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(g_seq_dat_w-1 DOWNTO 0);
  TYPE t_replicate_arr IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(c_dp_stream_data_w-1 DOWNTO 0);
  
  -- Registers in dp_clk domain
  SIGNAL ctrl_reg       : STD_LOGIC_VECTOR(c_mm_reg.nof_dat*c_mm_reg.dat_w-1 DOWNTO 0) := (OTHERS=>'0');
  
  SIGNAL diag_en        : STD_LOGIC;
  SIGNAL diag_sel       : STD_LOGIC;
  SIGNAL diag_init_mm   : STD_LOGIC_VECTOR(c_mm_reg.dat_w-1 DOWNTO 0) := (OTHERS=>'0');
  SIGNAL diag_init      : STD_LOGIC_VECTOR(g_seq_dat_w-1 DOWNTO 0) := (OTHERS=>'0');
  
  SIGNAL tx_dat_arr     : t_dat_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL tx_val_arr     : STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);
  SIGNAL tx_req_arr     : STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);

  SIGNAL tx_replicate_arr  : t_replicate_arr(g_nof_streams-1 DOWNTO 0);
  
BEGIN

  gen_nof_streams: FOR I IN 0 to g_nof_streams-1 GENERATE
    u_diag_tx_seq: ENTITY WORK.diag_tx_seq
    GENERIC MAP (
      g_dat_w    => g_seq_dat_w
    )
    PORT MAP (
      rst      => dp_rst,
      clk      => dp_clk,

      -- Write and read back registers:
      diag_en  => diag_en,
      diag_sel => diag_sel,
      diag_dat => diag_init,

      -- Streaming
      diag_req => tx_req_arr(I),
      out_dat  => tx_dat_arr(I),
      out_val  => tx_val_arr(I)
    );
    
    tx_req_arr(I) <= tx_src_in_arr(I).ready;
    
    
    tx_replicate_arr(I) <= REPLICATE_DP_DATA(tx_dat_arr(I));
    
    -- for some reason the intermediate tx_replicate_arr() signal is needed, otherwise the assignment to the tx_src_out_arr().data field remains void in the Wave window
    tx_src_out_arr(I).data  <= tx_replicate_arr(I);
    tx_src_out_arr(I).valid <= tx_val_arr(I);
  END GENERATE;

  -- Register mapping
  diag_en      <= ctrl_reg(                           0);  -- address 0, data bit [0]
  diag_sel     <= ctrl_reg(                           1);  -- address 0, data bit [1]
  diag_init_mm <= ctrl_reg(2*c_word_w-1 DOWNTO c_word_w);  -- address 1, data bits [31:0]
  
  diag_init    <= RESIZE_UVEC(diag_init_mm, g_seq_dat_w);

  u_reg : ENTITY common_lib.common_reg_r_w_dc
  GENERIC MAP (
    g_cross_clock_domain => TRUE,
    g_readback           => TRUE,  -- using TRUE fits for write and readback register
    g_reg                => c_mm_reg
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,
    st_rst      => dp_rst,
    st_clk      => dp_clk,
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in      => reg_mosi,
    sla_out     => reg_miso,
    
    -- MM registers in dp_clk domain
    in_reg      => ctrl_reg,  -- connect out_reg to in_reg for write and readback register
    out_reg     => ctrl_reg
  );
  
END str;













