--------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
-- Purpose: test bench for diag_block_gen 
--   The tb is self-stopping and self-checking.
-- Usage:
--   > as 5
--   > run -all

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL; 
USE common_lib.tb_common_pkg.ALL;     
USE common_lib.tb_common_mem_pkg.ALL; 
USE dp_lib.dp_stream_pkg.ALL;
USE dp_lib.tb_dp_pkg.ALL;
USE work.diag_pkg.ALL;

ENTITY tb_diag_block_gen IS
  GENERIC (
    g_buf_adr_w      : NATURAL := 7;   -- Waveform buffer address width (requires corresponding c_buf_file)
    g_buf_dat_w      : NATURAL := 32   -- Waveform buffer stored data width (requires corresponding c_buf_file)
  );
END tb_diag_block_gen;


ARCHITECTURE tb OF tb_diag_block_gen IS

  CONSTANT clk_period       : TIME    := 10 ns;
  CONSTANT c_timeout        : NATURAL := 1000;
  
  -- Default settings
  CONSTANT c_buf            : t_c_mem := (latency  => 1,
                                          adr_w    => g_buf_adr_w,
                                          dat_w    => g_buf_dat_w,     
                                          nof_dat  => 2**g_buf_adr_w,  -- = 2**adr_w
                                          init_sl  => '0');
                                          
  CONSTANT c_buf_file   : STRING := sel_a_b(c_buf.adr_w=7 AND c_buf.dat_w=32, "data/diag_block.hex", "UNUSED");
  
  CONSTANT c_cntr_init  : INTEGER := 0;
  CONSTANT c_cntr_incr  : INTEGER := 1;
  CONSTANT c_cntr_arr   : t_slv_32_arr(c_buf.nof_dat-1 DOWNTO 0) := flip(array_init(c_cntr_init, c_buf.nof_dat, c_cntr_incr));
                                             
  -- Default BG control
  CONSTANT c_bg_ctrl    : t_diag_block_gen := (        '0',      
                                                       '0',      
                                               TO_UVEC(96, c_diag_bg_samples_per_packet_w), 
                                               TO_UVEC(10, c_diag_bg_blocks_per_sync_w),      
                                               TO_UVEC(32, c_diag_bg_gapsize_w), 
                                               TO_UVEC( 0, c_diag_bg_mem_low_adrs_w),      
                                               TO_UVEC(95, c_diag_bg_mem_high_adrs_w), 
                                               TO_UVEC(42, c_diag_bg_bsn_init_w));
  CONSTANT c_bg_period  : NATURAL := TO_UINT(c_bg_ctrl.samples_per_packet) + TO_UINT(c_bg_ctrl.gapsize);
  
  -- Some alternative BG control settings
  CONSTANT c_more_samples_per_packet        : NATURAL := c_buf.nof_dat;
  CONSTANT c_less_samples_per_packet        : NATURAL := 5;
  CONSTANT c_alternative_mem_high_adrs      : NATURAL := 64;
  CONSTANT c_alternative_data_gap           : NATURAL := TO_UINT(c_bg_ctrl.mem_high_adrs)-c_alternative_mem_high_adrs;
                                               
  -- Another BG control for verifying XON
  CONSTANT c_bg_ctrl2   : t_diag_block_gen := (        '0',      
                                                       '0',      
                                               TO_UVEC(17, c_diag_bg_samples_per_packet_w), 
                                               TO_UVEC(10, c_diag_bg_blocks_per_sync_w),      
                                               TO_UVEC( 0, c_diag_bg_gapsize_w), 
                                               TO_UVEC( 0, c_diag_bg_mem_low_adrs_w),      
                                               TO_UVEC(16, c_diag_bg_mem_high_adrs_w), 
                                               TO_UVEC( 0, c_diag_bg_bsn_init_w));
                                               
  CONSTANT c_bg_period2 : NATURAL := TO_UINT(c_bg_ctrl2.samples_per_packet) + TO_UINT(c_bg_ctrl2.gapsize);
                                               
  SIGNAL tb_end         : STD_LOGIC := '0';
  SIGNAL rst            : STD_LOGIC;
  SIGNAL clk            : STD_LOGIC := '1';
  
  SIGNAL verify_en      : STD_LOGIC := '0';
  
  SIGNAL mm_buf_arr     : t_slv_32_arr(c_buf.nof_dat-1 DOWNTO 0);
  SIGNAL mm_buf_mosi    : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL mm_buf_miso    : t_mem_miso;
  
  SIGNAL bg_buf_mosi    : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL bg_buf_miso    : t_mem_miso;
  
  SIGNAL bg_ctrl        : t_diag_block_gen;
                                                                            
  SIGNAL out_siso       : t_dp_siso;
  SIGNAL out_sosi       : t_dp_sosi;
  SIGNAL prev_out_sosi  : t_dp_sosi;
  SIGNAL hold_sop       : STD_LOGIC;
  SIGNAL last_size      : NATURAL;
  SIGNAL exp_size       : NATURAL;
  SIGNAL cnt_size       : NATURAL;
  
BEGIN

  rst <= '1', '0' AFTER clk_period*10;

  clk <= NOT clk OR tb_end AFTER clk_period/2;
  
  p_stimuli : PROCESS
  BEGIN
    out_siso <= c_dp_siso_rdy;
    bg_ctrl <= c_bg_ctrl;
    proc_common_wait_until_low(clk, rst);       -- Wait until reset has finished
    proc_common_wait_some_cycles(clk, 10);
    
    -- Write waveform buffer
    proc_mem_write_ram(c_cntr_arr, clk, mm_buf_mosi);
    
    -- Readback waveform buffer
    proc_mem_read_ram(clk, mm_buf_mosi, mm_buf_miso, mm_buf_arr);
    ASSERT c_cntr_arr=mm_buf_arr REPORT "Error: Wrong BG buffer readback" SEVERITY FAILURE;
    
    -- Run with a gap
    bg_ctrl.enable <= '1';
    proc_dp_verify_run_some_cycles(5, 1500, 0, clk, verify_en);
    
    -- Run without gap
    bg_ctrl.enable <= '0';
    proc_common_wait_some_cycles(clk, 100); 
    bg_ctrl.gapsize <= TO_UVEC(0, c_diag_bg_gapsize_w);
    bg_ctrl.enable <= '1';
    proc_dp_verify_run_some_cycles(5, 1500, 0, clk, verify_en);
    
    -- Run with non-aligned memory settings
    bg_ctrl.enable <= '0';
    proc_common_wait_some_cycles(clk, 100);              
    bg_ctrl.gapsize <= c_bg_ctrl.gapsize;
    bg_ctrl.mem_high_adrs <= TO_UVEC(c_alternative_mem_high_adrs, c_diag_bg_mem_high_adrs_w);
    bg_ctrl.enable <= '1';
    proc_dp_verify_run_some_cycles(5, 1500, 0, clk, verify_en);
    
    -- Run and change bg_ctrl dynamically
    bg_ctrl.enable <= '0';
    proc_common_wait_some_cycles(clk, 100); 
    bg_ctrl <= c_bg_ctrl;
    bg_ctrl.enable <= '1';
    proc_dp_verify_run_some_cycles(5, 1500, 0, clk, verify_en);
    
    -- . change gapsize dynamically
    bg_ctrl.gapsize <= TO_UVEC(0, c_diag_bg_gapsize_w);
    proc_dp_verify_run_some_cycles(0, 1500, 0, clk, verify_en);

    bg_ctrl.gapsize <= c_bg_ctrl.gapsize;
    proc_dp_verify_run_some_cycles(0, 1500, 0, clk, verify_en);
    
    -- . change mem_high_adrs dynamically
    bg_ctrl.mem_high_adrs <= TO_UVEC(c_alternative_mem_high_adrs, c_diag_bg_mem_high_adrs_w);
    proc_dp_verify_run_some_cycles(5, 1500, 0, clk, verify_en);
    
    bg_ctrl.mem_high_adrs <= c_bg_ctrl.mem_high_adrs;
    proc_dp_verify_run_some_cycles(0, 1500, 0, clk, verify_en);
        
    -- . change samples_per_packet dynamically
    proc_common_wait_until_high(c_timeout, clk, out_sosi.sop);
    proc_common_wait_some_cycles(clk, 10);
    bg_ctrl.samples_per_packet <= TO_UVEC(c_more_samples_per_packet, c_diag_bg_samples_per_packet_w);   -- increase
    proc_dp_verify_run_some_cycles(0, 1500, 0, clk, verify_en);
    
    -- . change samples_per_packet dynamically
    proc_common_wait_until_high(c_timeout, clk, out_sosi.sop);
    proc_common_wait_some_cycles(clk, c_less_samples_per_packet*2);
    bg_ctrl.samples_per_packet <= TO_UVEC(c_less_samples_per_packet, c_diag_bg_samples_per_packet_w);   -- decrease, could yield proc_dp_verify_block_size() error if it occurs during a packet
    proc_dp_verify_run_some_cycles(0, 1500, 0, clk, verify_en);
    
    -- Run with XON flow control (verified by proc_dp_verify_block_size)
    bg_ctrl.enable <= '0';
    proc_common_wait_some_cycles(clk, 100); 
    bg_ctrl <= c_bg_ctrl2;
    bg_ctrl.enable <= '1';
    FOR I IN 0 TO c_bg_period2*3 LOOP
      out_siso <= c_dp_siso_rdy;
      proc_common_wait_some_cycles(clk, c_bg_period2*3+1);
      out_siso <= c_dp_siso_rst;
      proc_common_wait_some_cycles(clk, c_bg_period2*3);
    END LOOP;
    
    -- End simulation
    tb_end <= '1';
    WAIT;
  END PROCESS;
  
  -- Verification
  exp_size <= TO_UINT(bg_ctrl.samples_per_packet);
  last_size <= exp_size'last_value;
  
  proc_dp_verify_data("BSN",                                                            clk, verify_en, out_sosi.sop,   out_sosi.bsn,  prev_out_sosi.bsn);
  proc_dp_verify_data("DATA", TO_UINT(c_bg_ctrl.mem_high_adrs), c_alternative_data_gap, clk, verify_en, out_sosi.valid, out_sosi.data, prev_out_sosi.data);
  
  proc_dp_verify_sop_and_eop(clk, out_sosi.valid, out_sosi.sop, out_sosi.eop, hold_sop);
  proc_dp_verify_block_size(last_size, exp_size, clk, out_sosi.valid, out_sosi.sop, out_sosi.eop, cnt_size);

  -- Waveform buffer
  u_buf : ENTITY common_lib.common_ram_crw_crw
  GENERIC MAP (
    g_ram       => c_buf,
    g_init_file => c_buf_file
  )
  PORT MAP (
    rst_a     => '0',
    rst_b     => '0',
    clk_a     => clk,
    clk_b     => clk,
    wr_en_a   => mm_buf_mosi.wr,
    wr_en_b   => '0',
    wr_dat_a  => mm_buf_mosi.wrdata(c_buf.dat_w-1 DOWNTO 0),
    wr_dat_b  => (OTHERS=>'0'),
    adr_a     => mm_buf_mosi.address(c_buf.adr_w-1 DOWNTO 0),
    adr_b     => bg_buf_mosi.address(c_buf.adr_w-1 DOWNTO 0),
    rd_en_a   => mm_buf_mosi.rd,
    rd_en_b   => bg_buf_mosi.rd,
    rd_dat_a  => mm_buf_miso.rddata(c_buf.dat_w-1 DOWNTO 0),
    rd_dat_b  => bg_buf_miso.rddata(c_buf.dat_w-1 DOWNTO 0),
    rd_val_a  => mm_buf_miso.rdval,
    rd_val_b  => bg_buf_miso.rdval
  );
  
  u_dut : ENTITY work.diag_block_gen
  GENERIC MAP(
    g_buf_dat_w  => c_buf.dat_w,   
    g_buf_addr_w => c_buf.adr_w
  )             
  PORT MAP (         
    rst         => rst,
    clk         => clk,
    buf_addr    => bg_buf_mosi.address(c_buf.adr_w-1 DOWNTO 0), 
    buf_rden    => bg_buf_mosi.rd, 
    buf_rddat   => bg_buf_miso.rddata(c_buf.dat_w-1 DOWNTO 0),   
    buf_rdval   => bg_buf_miso.rdval,   
    ctrl        => bg_ctrl,
    out_siso    => out_siso,
    out_sosi    => out_sosi
  );
   
END tb;

