-----------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Test bench for mms_diag_block_gen
--   The tb is self-stopping and self-checking.
-- Usage:
--   > do wave_mms_diag_block_gen.do
--   > run -all
--   Observe tb_state and check the out_sosi fields if data is OK.
--   Try both c_gap_size = 0 and 160

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;   
USE IEEE.numeric_std.ALL;
USE std.textio.all;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL; 
USE dp_lib.dp_stream_pkg.ALL;
USE dp_lib.tb_dp_pkg.ALL;
USE work.diag_pkg.ALL;

ENTITY tb_mms_diag_block_gen IS
  GENERIC (
    g_nof_inputs       : POSITIVE := 64;                        
    g_nof_subbands     : POSITIVE := 24;
    g_nof_streams      : NATURAL  := 16;
    g_buf_dat_w        : NATURAL  := 32; 
    g_buf_addr_w       : NATURAL  := 7;
    g_file_name_prefix : STRING   := "data/bf_in_data"; 
    g_sim              : BOOLEAN  := TRUE
  );
END tb_mms_diag_block_gen;

ARCHITECTURE tb OF tb_mms_diag_block_gen IS
  
  CONSTANT clk_period    : TIME := 10 ns;
  
  CONSTANT c_block_size  : NATURAL := 96;
  --CONSTANT c_gap_size    : NATURAL := 0;
  CONSTANT c_gap_size    : NATURAL := 160;
  CONSTANT c_period_size : NATURAL := c_block_size + c_gap_size;
  
  --CONSTANT c_blk_sync    : BOOLEAN := FALSE;
  CONSTANT c_blk_sync    : BOOLEAN := TRUE;
  
  CONSTANT c_pend_period : NATURAL := 200;
  CONSTANT c_on_period   : NATURAL := 1500;
  CONSTANT c_off_period  : NATURAL := 550;
  CONSTANT c_end_period  : NATURAL := 5000;
  
  CONSTANT c_bsn_max     : UNSIGNED(c_dp_stream_bsn_w-1 DOWNTO 0) := (OTHERS=>'1');
  CONSTANT c_bsn_gap     : UNSIGNED(c_dp_stream_bsn_w-1 DOWNTO 0) := TO_UNSIGNED(ceil_div(c_off_period, c_period_size), c_dp_stream_bsn_w);   -- depends on c_off_period
  
  TYPE t_tb_state IS (s_init, s_enable_sync, s_enable, s_disable, s_xoff, s_xon);

  SIGNAL tb_end           : STD_LOGIC := '0';
  SIGNAL rst              : STD_LOGIC;
  SIGNAL clk              : STD_LOGIC := '1';
  SIGNAL en_sync          : STD_LOGIC;
  SIGNAL tb_state         : t_tb_state;
                         
  SIGNAL ram_bg_data_mosi : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL ram_bg_data_miso : t_mem_miso := c_mem_miso_rst;
  SIGNAL reg_bg_ctrl_mosi : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL reg_bg_ctrl_miso : t_mem_miso := c_mem_miso_rst;

  SIGNAL out_siso_arr     : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
  SIGNAL out_sosi_arr     : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);   
  
  SIGNAL verify_en        : STD_LOGIC := '1';
  SIGNAL out_sosi         : t_dp_sosi;
  SIGNAL prev_out_sosi    : t_dp_sosi;
  
BEGIN
  
  clk <= NOT clk OR tb_end AFTER clk_period/2;
  rst <= '1', '0' AFTER clk_period*5;
  
  p_write_bg_ctrl : PROCESS
  BEGIN
    en_sync <= '0';
    out_siso_arr <= (OTHERS=>c_dp_siso_rdy);                   -- Flow control XON
    
    -- BG init
    tb_state <= s_init;
    proc_common_wait_until_low(clk, rst);                      -- Wait until reset has finished
    proc_common_wait_some_cycles(clk, 10);                     -- Wait an additional amount of cycles
    proc_mem_mm_bus_wr(1, c_block_size, clk, reg_bg_ctrl_mosi);  -- Set the nof samples per block
    proc_mem_mm_bus_wr(2,           10, clk, reg_bg_ctrl_mosi);  -- Set the nof blocks per sync
    proc_mem_mm_bus_wr(3,   c_gap_size, clk, reg_bg_ctrl_mosi);  -- Set the gapsize
    proc_mem_mm_bus_wr(4,            0, clk, reg_bg_ctrl_mosi);  -- Set the mem low address
    proc_mem_mm_bus_wr(5,           95, clk, reg_bg_ctrl_mosi);  -- Set the mem high address
    proc_mem_mm_bus_wr(6,           42, clk, reg_bg_ctrl_mosi);  -- Set the lower part of the initial bsn
    proc_mem_mm_bus_wr(7,            0, clk, reg_bg_ctrl_mosi);  -- Set the higher part of the initial bsn
    
    -- BG enable at en_sync pulse
    proc_mem_mm_bus_wr(0,          3, clk, reg_bg_ctrl_mosi);  -- Enable the block generator at en_sync pulse. 
    tb_state <= s_enable_sync;
    proc_common_wait_some_cycles(clk, c_pend_period);          -- Wait an additional amount of cycles
    proc_common_gen_pulse(clk, en_sync);                       -- Issue the en_sync pulse to start the enabled block generator
    proc_common_wait_some_cycles(clk, c_on_period);            -- On time
    proc_common_wait_until_high(clk, out_sosi_arr(0).valid);   -- Wait until block generator is active
    proc_common_wait_some_cycles(clk, 20);                     -- Wait an additional amount of cycles
    
    -- BG disable
    proc_mem_mm_bus_wr(0,   0, clk, reg_bg_ctrl_mosi);         -- Disable the block generator. 
    tb_state <= s_disable;
    proc_common_wait_some_cycles(clk, c_off_period);           -- Off time
    
    -- BG enable
    proc_mem_mm_bus_wr(0,   1, clk, reg_bg_ctrl_mosi);         -- Enable the block generator. 
    tb_state <= s_enable;
    proc_common_wait_some_cycles(clk, c_on_period);            -- On time
    proc_common_wait_until_high(clk, out_sosi_arr(0).valid);   -- Wait until block generator is active
    proc_common_wait_some_cycles(clk, 20);                     -- Wait an additional amount of cycles
    
    -- BG flow off
    out_siso_arr <= (OTHERS=>c_dp_siso_rst);                   -- Flow control XOFF
    tb_state <= s_xoff;
    proc_common_wait_some_cycles(clk, c_off_period);           -- Off time
    
    -- BG flow on
    out_siso_arr <= (OTHERS=>c_dp_siso_rdy);                   -- Flow control XON
    tb_state <= s_xon;
    proc_common_wait_some_cycles(clk, c_end_period);           -- On time
    tb_end <= '1';
    WAIT;
  END PROCESS;
      
  -------------------------------------------------
  -- Instantiation of the device under test
  -------------------------------------------------
  u_dut : ENTITY work.mms_diag_block_gen
  GENERIC MAP (    
    g_blk_sync           => c_blk_sync,
    g_nof_output_streams => g_nof_streams, 
    g_buf_dat_w          => g_buf_dat_w,       
    g_buf_addr_w         => g_buf_addr_w,  
    g_file_name_prefix   => g_file_name_prefix     
  )
  PORT MAP (
    -- System
    mm_rst           => rst, 
    mm_clk           => clk, 
    dp_rst           => rst, 
    dp_clk           => clk, 
    en_sync          => en_sync,
    -- MM interface
    ram_bg_data_mosi => ram_bg_data_mosi,  
    ram_bg_data_miso => ram_bg_data_miso,  
    reg_bg_ctrl_mosi => reg_bg_ctrl_mosi,  
    reg_bg_ctrl_miso => reg_bg_ctrl_miso,  
    -- ST interface
    out_siso_arr     => out_siso_arr,
    out_sosi_arr     => out_sosi_arr
  );

  -------------------------------------------------
  -- Verification
  -------------------------------------------------
  
  p_compare_out_sosi_arr : PROCESS(clk)
  BEGIN
    IF rising_edge(clk) THEN
      IF rst='0' THEN
        FOR I IN 0 TO g_nof_streams-1 LOOP
          -- All outputs have the same framing
          ASSERT out_sosi_arr(0).sync =out_sosi_arr(I).sync  REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).sync";
          ASSERT out_sosi_arr(0).bsn  =out_sosi_arr(I).bsn   REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).bsn";
          ASSERT out_sosi_arr(0).im   =out_sosi_arr(I).im    REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).im";
          ASSERT out_sosi_arr(0).valid=out_sosi_arr(I).valid REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).valid";
          ASSERT out_sosi_arr(0).sop  =out_sosi_arr(I).sop   REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).sop";
          ASSERT out_sosi_arr(0).eop  =out_sosi_arr(I).eop   REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).eop";
          
          -- Only the re part differs per output given the BG buffer data from g_file_name_prefix
          ASSERT TO_UINT(out_sosi_arr(0).re)=TO_UINT(out_sosi_arr(I).re)-I*4 REPORT "tb_mms_diag_block_gen: Wrong out_sosi_arr(*).re";        
        END LOOP;
      END IF;
    END IF;
  END PROCESS;
  
  out_sosi <= out_sosi_arr(0);
  
  p_verify_en : PROCESS
  BEGIN
    -- t_tb_state = s_init, s_enable_sync, s_enable
    verify_en <= '0';
    proc_common_wait_some_cycles(clk, 400);
    verify_en <= '1';
    proc_common_wait_some_cycles(clk, c_on_period);
    -- t_tb_state = s_disable
    verify_en <= '0';
    proc_common_wait_some_cycles(clk, c_off_period);
    -- t_tb_state = s_xoff, s_xon
    verify_en <= '1';  -- use c_bsn_gap to cover gap in BSN during s_xoff
    WAIT;
  END PROCESS;
    
  proc_dp_verify_data("BSN", c_bsn_max, c_bsn_gap, clk, verify_en, out_sosi.sop, out_sosi.bsn, prev_out_sosi.bsn);
  
END tb;