-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.common_network_total_header_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.eth_pkg.ALL;

-- Purpose:
--   Determine the Ethernet header status.
-- Description:
--   Determines all relevant information from the Ethernet header after every
--   asserted sop.
-- Remark:
-- . g_ip_header_checksum_calculate:
--   = TRUE, then the IP header checksum is calculated for hdr_status
--   = FALSE, then the IP header checksum is read for hdr_status

ENTITY eth_hdr_status IS
  GENERIC (
    g_ip_header_checksum_calculate : BOOLEAN := TRUE
  );
  PORT (
    rst               : IN  STD_LOGIC;
    clk               : IN  STD_LOGIC;
    
    -- Total header
    hdr_words_arr     : IN  t_network_total_header_32b_arr;
    hdr_words_arr_val : IN  STD_LOGIC_VECTOR(0 TO c_network_total_header_32b_nof_words-1);
    hdr_fields        : IN  t_network_total_header;
    hdr_fields_val    : IN  STD_LOGIC_VECTOR(0 TO c_network_total_header_32b_nof_words-1);
    hdr_data          : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    hdr_data_val      : IN  STD_LOGIC;
    
    -- Header status
    hdr_status        : OUT t_eth_hdr_status
  );
END eth_hdr_status;


ARCHITECTURE rtl OF eth_hdr_status IS

  SIGNAL i_hdr_status        : t_eth_hdr_status;
  SIGNAL nxt_hdr_status      : t_eth_hdr_status;
  
  SIGNAL ip_header           : t_dp_sosi;
  SIGNAL calc_checksum       : STD_LOGIC_VECTOR(c_halfword_w-1 DOWNTO 0);
  SIGNAL calc_checksum_val   : STD_LOGIC;
  SIGNAL ip_checksum         : STD_LOGIC_VECTOR(c_halfword_w-1 DOWNTO 0);
  SIGNAL ip_checksum_val     : STD_LOGIC;
  
BEGIN

  hdr_status <= i_hdr_status;
  
  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      i_hdr_status <= c_eth_hdr_status_rst;
    ELSIF rising_edge(clk) THEN
      i_hdr_status <= nxt_hdr_status;
    END IF;
  END PROCESS;
  
  ------------------------------------------------------------------------------
  -- Packet type
  nxt_hdr_status.is_arp  <= '1' WHEN UNSIGNED(hdr_fields.eth.eth_type)=c_network_eth_type_arp     AND hdr_fields_val(c_network_total_header_32b_eth_type_wi)='1' ELSE '0';
  nxt_hdr_status.is_ip   <= '1' WHEN UNSIGNED(hdr_fields.eth.eth_type)=c_network_eth_type_ip      AND hdr_fields_val(c_network_total_header_32b_eth_type_wi)='1' ELSE '0';
  nxt_hdr_status.is_icmp <= '1' WHEN UNSIGNED(hdr_fields.ip.protocol)=c_network_ip_protocol_icmp  AND hdr_fields_val(c_network_total_header_32b_ip_protocol_wi)='1'  AND i_hdr_status.is_ip='1'  ELSE '0';
  nxt_hdr_status.is_udp  <= '1' WHEN UNSIGNED(hdr_fields.ip.protocol)=c_network_ip_protocol_udp   AND hdr_fields_val(c_network_total_header_32b_ip_protocol_wi)='1'  AND i_hdr_status.is_ip='1'  ELSE '0';
  nxt_hdr_status.is_dhcp <= '1' WHEN UNSIGNED(hdr_fields.udp.dst_port)=c_network_udp_port_dhcp_in AND hdr_fields_val(c_network_total_header_32b_udp_dst_port_wi)='1' AND i_hdr_status.is_udp='1' ELSE '0';
  
  ------------------------------------------------------------------------------
  -- IP Header checksum
  ip_header.data  <= RESIZE_dp_data(hdr_data);
  ip_header.valid <= '1' WHEN UNSIGNED(hdr_words_arr_val(c_network_total_header_32b_ip_lo_wi TO c_network_total_header_32b_ip_hi_wi))/=0 ELSE '0';
  ip_header.sop   <= hdr_words_arr_val(c_network_total_header_32b_ip_lo_wi);
  ip_header.eop   <= hdr_words_arr_val(c_network_total_header_32b_ip_hi_wi);
  ip_header.empty <= (OTHERS=>'0');
  
  -- Calculate the IP header checksum for the packet header
  u_calc_checksum : ENTITY work.eth_checksum
  PORT MAP (
    rst           => rst,
    clk           => clk,
    snk_in        => ip_header,
    checksum      => calc_checksum,
    checksum_val  => calc_checksum_val
  );
  
  -- Use calculated IP header checksum
  gen_calc : IF g_ip_header_checksum_calculate=TRUE GENERATE
    ip_checksum     <= calc_checksum;
    ip_checksum_val <= calc_checksum_val;
  END GENERATE;
  
  -- Use IP header checksum field read from the frame header
  gen_read : IF g_ip_header_checksum_calculate=FALSE GENERATE
    ip_checksum     <= hdr_fields.ip.header_checksum;
    ip_checksum_val <= calc_checksum_val;              -- use u_calc_checksum only for valid control
  END GENERATE;
  
  -- If it is an IP packet then update the hdr_status IP checksum fields, else leave them 0
  p_hdr_status_ip_checksum : PROCESS(i_hdr_status, ip_checksum, ip_checksum_val)
  BEGIN
    nxt_hdr_status.ip_checksum_val   <= '0';
    nxt_hdr_status.ip_checksum       <= (OTHERS=>'0');
    nxt_hdr_status.ip_checksum_is_ok <= '0';
    IF i_hdr_status.is_ip='1' AND ip_checksum_val='1' THEN
      nxt_hdr_status.ip_checksum_val <= '1';
      nxt_hdr_status.ip_checksum     <= ip_checksum;
      IF UNSIGNED(ip_checksum)=0 THEN
        nxt_hdr_status.ip_checksum_is_ok <= '1';
      END IF;
    END IF;
  END PROCESS;
    
  ------------------------------------------------------------------------------
  -- UDP port number
  nxt_hdr_status.udp_port <= hdr_fields.udp.dst_port;
  
END rtl;
