--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------


-- Purpose: Combine mac_10g and 10gbase_r for c_tech_arria10
-- Description 
--   
--   The clocks come from an external central fPLL:
--
--     tx_ref_clk_644 --> fPLL --> clk_312
--                                 clk_156, rst_156
--   Blockdiagram:
--
--                          312 156 644  
--                _________   |  |  |   ____________
--                |       |   |  |  |   |          |
--                |       |<--/  |  \-->|          |
--                |       |<-----+----->|          |
--                |       |             |          |
--                |       |    XGMII    |          |
--     tx_snk --->|tech_  |------------>|tech_     |---> serial_tx
--     rx_src <---|mac_10g|<------------|10gbase_r |<--- serial_rx
--                |       |             |          |
--                |_______|--\       /--|__________|
--                    |      |       |
--                  mac_mm   |       |
--                           |       v
--                       (   v    xgmii_tx_ready)
--     tx_snk_out.xon <--(xgmii_link_status[1:0])
--
--  When the xgmii_tx_ready from the 10gbase_r and the xgmii_link_status from
--  the mac_10g are both be OK then the tx_snk.xon is asserted to allow the
--  user data transmission.
--  The tb_tech_eth_10g reveals that xgmii_tx_ready goes high after some power
--  up time and then remains active independent of link_fault. A link fault
--  eg. due to rx disconnect is detected by xgmii_link_status:
--    0 = OK
--    1 = local fault
--    2 = remote fault
--  Hence when the xgmii_link_status is OK then the other side is also OK so
--  then it is also appropriate to release tx_snk.xon.

LIBRARY IEEE, common_lib, dp_lib, technology_lib, tech_10gbase_r_lib, tech_mac_10g_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE technology_lib.technology_pkg.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_mac_10g_lib.tech_mac_10g_component_pkg.ALL;

ENTITY tech_eth_10g_arria10 IS
  GENERIC (
    g_sim                 : BOOLEAN := FALSE;
    g_sim_level           : NATURAL := 0;     -- 0 = use IP; 1 = use fast serdes model
    g_nof_channels        : NATURAL := 1;
    g_link_status_check   : STD_LOGIC_VECTOR(c_tech_mac_10g_link_status_w-1 DOWNTO 0) := "11";
    g_pre_header_padding  : BOOLEAN := FALSE
  );
  PORT (
    -- Transceiver PLL reference clock
    tr_ref_clk_644   : IN  STD_LOGIC := '0';   -- 644.531250 MHz for 10GBASE-R
    
    -- Data clocks
    clk_312          : IN  STD_LOGIC := '0';
    clk_156          : IN  STD_LOGIC := '0';
    rst_156          : IN  STD_LOGIC := '0';
    
    -- MM
    mm_clk           : IN  STD_LOGIC;
    mm_rst           : IN  STD_LOGIC;
    
    mac_mosi         : IN  t_mem_mosi;         -- MAG_10G (CSR)
    mac_miso         : OUT t_mem_miso; 
    
    -- ST
    tx_snk_in_arr    : IN  t_dp_sosi_arr(g_nof_channels-1 DOWNTO 0);      -- 64 bit data @ clk_156
    tx_snk_out_arr   : OUT t_dp_siso_arr(g_nof_channels-1 DOWNTO 0); 
    
    rx_src_out_arr   : OUT t_dp_sosi_arr(g_nof_channels-1 DOWNTO 0);      -- 64 bit data @ clk_156
    rx_src_in_arr    : IN  t_dp_siso_arr(g_nof_channels-1 DOWNTO 0);
    
    -- Serial
    serial_tx_arr    : OUT STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);
    serial_rx_arr    : IN  STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0)
  );
END tech_eth_10g_arria10;


ARCHITECTURE str OF tech_eth_10g_arria10 IS

  -- MAG_10G control status registers
  SIGNAL mac_mosi_arr          : t_mem_mosi_arr(g_nof_channels-1 DOWNTO 0);
  SIGNAL mac_miso_arr          : t_mem_miso_arr(g_nof_channels-1 DOWNTO 0); 

  -- XON control
  SIGNAL mac_snk_out_arr       : t_dp_siso_arr(g_nof_channels-1 DOWNTO 0);
  
  -- XGMII
  SIGNAL xgmii_link_status_arr : t_tech_mac_10g_xgmii_status_arr(g_nof_channels-1 DOWNTO 0);  -- 2 bit, from MAC_10g
  SIGNAL xgmii_tx_ready_arr    : STD_LOGIC_VECTOR(g_nof_channels-1 DOWNTO 0);                 -- 1 bit, from PHY 10gbase_r
  SIGNAL xgmii_tx_dc_arr       : t_xgmii_dc_arr(g_nof_channels-1 DOWNTO 0);    -- 72 bit
  SIGNAL xgmii_rx_dc_arr       : t_xgmii_dc_arr(g_nof_channels-1 DOWNTO 0);    -- 72 bit

BEGIN
                   
  gen_mac : FOR I IN 0 TO g_nof_channels-1 GENERATE
  
    tx_snk_out_arr(I).ready <= mac_snk_out_arr(I).ready;  -- pass on MAC cycle accurate backpressure
    
    p_xon_flow_control : PROCESS(clk_156)
      VARIABLE v_xgmii_link_status : STD_LOGIC_VECTOR(c_tech_mac_10g_link_status_w-1 DOWNTO 0);
    BEGIN
      IF rising_edge(clk_156) THEN
        tx_snk_out_arr(I).xon <= '0';
        v_xgmii_link_status := xgmii_link_status_arr(I) AND g_link_status_check;  -- use mask to check Tx, Rx, both or none. 
        IF xgmii_tx_ready_arr(I)='1' AND v_xgmii_link_status="00" THEN
          tx_snk_out_arr(I).xon <= '1';  -- XON when Tx PHY is ready and XGMII is ok
        END IF;
      END IF;
    END PROCESS;
    
    u_tech_mac_10g : ENTITY tech_mac_10g_lib.tech_mac_10g
    GENERIC MAP (
      g_technology          => c_tech_arria10,
      g_pre_header_padding  => g_pre_header_padding
    )
    PORT MAP (
      -- MM
      mm_clk            => mm_clk,
      mm_rst            => mm_rst,
      csr_mosi          => mac_mosi_arr(I),
      csr_miso          => mac_miso_arr(I),
  
      -- ST
      tx_clk_312        => clk_312,
      tx_clk_156        => clk_156,
      tx_rst            => rst_156,
      tx_snk_in         => tx_snk_in_arr(I),   -- 64 bit data
      tx_snk_out        => mac_snk_out_arr(I),
      
      rx_clk_312        => clk_312,
      rx_clk_156        => clk_156,
      rx_rst            => rst_156,
      rx_src_out        => rx_src_out_arr(I),  -- 64 bit data
      rx_src_in         => rx_src_in_arr(I),
            
      -- XGMII
      xgmii_link_status => xgmii_link_status_arr(I),
      xgmii_tx_data     => xgmii_tx_dc_arr(I),
      xgmii_rx_data     => xgmii_rx_dc_arr(I)
    );
  END GENERATE;  
  
  u_tech_10gbase_r: ENTITY tech_10gbase_r_lib.tech_10gbase_r
  GENERIC MAP (
    g_technology     => c_tech_arria10,
    g_sim            => g_sim,
    g_sim_level      => g_sim_level,
    g_nof_channels   => g_nof_channels
  )
  PORT MAP (
    -- Transceiver PLL reference clock
    tr_ref_clk_644     => tr_ref_clk_644,
    
    -- XGMII clocks
    clk_156            => clk_156,
    rst_156            => rst_156,

    -- XGMII interface
    xgmii_tx_ready_arr => xgmii_tx_ready_arr,
    xgmii_rx_ready_arr => OPEN,
    xgmii_tx_dc_arr    => xgmii_tx_dc_arr,
    xgmii_rx_dc_arr    => xgmii_rx_dc_arr,

    -- PHY serial IO
    tx_serial_arr      => serial_tx_arr,
    rx_serial_arr      => serial_rx_arr
  );
    
  -----------------------------------------------------------------------------
  -- MM bus mux
  -----------------------------------------------------------------------------
  u_common_mem_mux : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_channels,
    g_mult_addr_w => func_tech_mac_10g_csr_addr_w(c_tech_arria10) 
  )
  PORT MAP (
    mosi     => mac_mosi,
    miso     => mac_miso,
    mosi_arr => mac_mosi_arr,
    miso_arr => mac_miso_arr
  );  

END str;
